// Payload: An LMM/XMM Program Loader for Catalina. This program can be used
//          to load SPIN or LMM programs. In combination with an XMM loader
//          it can be also used to load XMM programs. Normal SPIN or LMM
//          programs are loaded by specifying them as the first (and only) 
//          file - e.g:
//
//             payload  my_SPIN_or_LMM_program.binary
//
//          XMM programs must be loaded by specifying an XMM loader program
//          as the first file, and the program to be loaded as the second
//          file - e.g:
//
//             payload  XMM_LOADER.binary  my_XMM_program.binary
//           
// This version by Ross Higson, partly based on biffprop.c by bifferos, which
// was itself partly based on Loader.py by Remy Blank. 
//
// ------------------------------------------------------------------------
// Caution: This version of payload is  an experimental version modified
// by Andrew Robinson Dec 20,2010 
// The primary reason for modifying it was that the original payload
// malfunctioned on a test 64 bit Ubuntu platform
// The present version is a beta modification with the following properties:
// It works on Ubuntu 10.10 with an Intel core duo (64 bit) at *FULL SPEED*
// It has an advanced packing algorithm that increases the data upload rate
// and detection of the Propeller platform -- on the down side, it may pollute
// the serial line after loading the propeller with a short string of 
// irrelevant data before releasing the line; but this method of detecting
// responses ought to be less sensitive to system timing variations between
// un*x platforms of various kinds.
//
// Slower baudrates have difficulties which I am uncertain of the origin of,
// However, on another version of payload --
// Ross H -- modified the nice values (on 32 bit versions of Ubuntu) which were
// also causing problems, and was able to get his version to function that way
// So that may also be a problem here
// On my version of Ubuntu, starting a communications program will reset the
// propeller; so for others having the same problem, look for a modified
// version of nanocom that can stay running while payload uploads a program
// One such version is on the parallax website posted by yours truly.
//
// --Andrew
// ------------------------------------------------------------------------
//
//
// Version 2.4 - Initial release
//
// Version 2.5 - Add EEPROM programming option (-e).
//               Print message on error responses (load, program or verify).  
//
// Version 2.6 - Just update version number
//
// Version 2.7 - just update version number.
//
//-----------------------------------------------------------------------------
// Payload is part of Catalina.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// Free Software Foundation, version 2.
//
// This program is distributed in the hope that it will be useful, but 
// WITHOUT ANY WARRANTY; without even the implied warranty of 
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
// Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
// ----------------------------------------------------------------------------

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>
#include <errno.h>

#include "rs232.h"

#define VERSION            "2.7"

#define MAX_FILES          3 // 3 is required only for cpu-to-prop-to-prop loads (using intermediate loader)
#define MAX_LINELEN        1000

#define MAX_PROGRAM_SIZE   32768 // for LMM and SPIN programs, not XMM programs

#define CMD_SHUTDOWN       0
#define CMD_LOADRUN        1  // load into RAM
#define CMD_PROGRUN        3  // program into EEPROM

#define MIN_BAUDRATE       19200
#define MAX_BAUDRATE       3000000
#define DEFAULT_BAUDRATE   115200

#define MAX_CPU            255
#define DEFAULT_CPU        1

#define DEFAULT_TIMEOUT    250 // milliseconds
#define DEFAULT_SYNCTIME   100 // milliseconds
#define MAX_TIMEOUT        10000 // milliseconds

#define RESET_TIME         15 // milliseconds
#define MAX_ATTEMPTS       5

#define PAGE_SIZE          512

#define SIO_EOP            0x00FFFFFE

#define INTERCHAR_DELAY    0 // milliseconds
#define INTERPAGE_DELAY    25 // milliseconds

#ifdef __linux__   /* Linux */

// these have been set by trial and error, and may vary on different machines

#define RESET_DELAY        0 // milliseconds
#define INTERFILE_DELAY    1000 // milliseconds

#else              /* Windows */

// these have been set by trial and error, and may vary on different machines

#define RESET_DELAY        0 // milliseconds
#define INTERFILE_DELAY    500 // milliseconds

#endif

typedef unsigned long u32;
typedef unsigned char u8;


static int    auto_port = 1;  // 1 means auto (starting at port 0)
static int    port = 0;
static int    eeprom = 0;
static int    cpu = DEFAULT_CPU;
static int    baudrate = DEFAULT_BAUDRATE;
static int    timeout = DEFAULT_TIMEOUT;
static int    synctime = DEFAULT_SYNCTIME;
static int    wait = 0;
static int    xmm_only = 0;
static int    double_reset = 0;
static int    reset_delay = RESET_DELAY;
static int    interfile_delay = INTERFILE_DELAY;
static int    second_port = -1; // -1 => use first port
static int    verbose = 0;
static int    diagnose = 0;
static int    file_count = 0;                  
static char * file_name[MAX_FILES];


// safecpy will never write more than size characters, 
// and is guaranteed to null terminate its result, so
// make sure the buffer passed is at least size + 1
char * safecpy(char *dst, const char *src, size_t size) {
   dst[size] = '\0';
   if (src) {
      return strncpy(dst, src, size - strlen(dst));
   }
}


// safecat will never write more than size characters, 
// and is guaranteed to null terminate its resul, so
// make sure the buffer passed is at least size + 1
char * safecat(char *dst, const char *src, size_t size) {
   dst[size] = '\0';
   if (src) {
      return strncat(dst, src, size - strlen(dst));
   }
}

inline
void ESendByte( int port, u8 data ){
	if ( SendByte(port,data) < 0 ) {
		perror("Sending of serial data failed:");
		exit(1);
	}
}

// A single output stream that can be printed as a block to detect propellers
unsigned char lfsrOut[251]={
0xCF,0xDE,0xDE,0xDF,0xCE,0xDF,0xDF,0xDE,0xDE,0xDF,0xDF,0xCE,0xDE,0xDF,0xCE,0xCF,
0xCF,0xCE,0xDF,0xDF,0xCE,0xCE,0xCF,0xDE,0xCE,0xCF,0xDF,0xDF,0xCE,0xCF,0xCE,0xCF,
0xCE,0xDF,0xDE,0xCF,0xDF,0xCF,0xCE,0xCF,0xDF,0xCE,0xDF,0xCE,0xCF,0xDF,0xDE,0xCF,
0xDE,0xCE,0xDF,0xCF,0xCF,0xCF,0xCF,0xDF,0xCF,0xDF,0xDE,0xDE,0xCF,0xCF,0xDE,0xDF,
0xDF,0xDF,0xCF,0xDE,0xCF,0xDF,0xDF,0xDE,0xDF,0xDE,0xCE,0xCE,0xCE,0xDE,0xDE,0xDE,
0xCF,0xCE,0xDF,0xCE,0xDF,0xCF,0xCE,0xCE,0xDE,0xDF,0xDF,0xDE,0xCE,0xCF,0xCF,0xCF,
0xDE,0xCE,0xCE,0xDE,0xCE,0xCE,0xDF,0xCF,0xDF,0xDF,0xDF,0xDE,0xCF,0xCE,0xDE,0xCF,
0xCE,0xCF,0xDE,0xCF,0xCE,0xCE,0xDF,0xDE,0xCE,0xDE,0xDE,0xCE,0xDF,0xDE,
// Reader probe bytes follow; each extracts 2 bits of LFSR data
0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,
0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,
0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,
0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,
0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,
0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,
0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,
0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5,0xA5
};

// This is the expected reply signature of a Propeller chip to compare against
unsigned char lfsrIn[125]={
0xE6,0xC6,0xC6,0xC7,0xE7,0xC7,0xE6,0xE7,0xC7,0xC7,0xE7,0xE7,0xC7,0xC6,0xE7,0xC7,
0xE6,0xE6,0xC6,0xE6,0xE7,0xC7,0xC6,0xE6,0xC6,0xC7,0xE6,0xE6,0xE7,0xC7,0xE6,0xC6,
0xE6,0xC6,0xE6,0xC7,0xE7,0xE6,0xE7,0xC6,0xE6,0xE6,0xC7,0xE6,0xC7,0xE6,0xE6,0xC7,
0xE7,0xC6,0xC7,0xE6,0xE7,0xE6,0xE6,0xE6,0xE6,0xE7,0xE6,0xC7,0xC7,0xE7,0xE6,0xC6,
0xE7,0xE7,0xE7,0xE7,0xC6,0xE7,0xE6,0xE7,0xC7,0xE7,0xC7,0xC7,0xC6,0xC6,0xC6,0xC7,
0xC7,0xE7,0xC6,0xE6,0xC7,0xE6,0xE7,0xC6,0xC6,0xC6,0xE7,0xE7,0xC7,0xC7,0xE6,0xE6,
0xE6,0xC6,0xC7,0xC6,0xC6,0xC7,0xC6,0xE6,0xE7,0xE6,0xE7,0xE7,0xC7,0xE7,0xC6,0xC6,
0xE7,0xC6,0xE6,0xC6,0xE7,0xC6,0xC6,0xE6,0xC7,0xC7,0xC6,0xC7,0xC7
};

void Send2Bits( int port, char data ) {
	data = 0xCE | (data&1) | (data&2)<<3; // Two bit packing
	ESendByte( port, data );
}

int Get2Unaligned( int port ) {
	unsigned i0, i1;
	int ch;
	errno=0;
	ESendByte( port, 0xA5 ); // 10100:101_  probe for 2 bits (F,T->Q,F->T,T)
	if (((ch = ReceiveByte( port ))<0) || errno ) {
		if (diagnose) perror("Get2Unaligned");
		fprintf(stderr,"ch=%02X\n", (unsigned)ch );
		return -1;
	}
	if (errno) return -1; // any errors

	i0=!!(ch & 0x01); // 11x0011?	<-- Reception bits are located at '?'
	i1=!!(ch & 0x20); // 11?0011x
	if ((ch & 0xDE)!=0xC6) {
		if (diagnose) 	fprintf(stderr,"Invalid reply: %02X\n",
					(unsigned)ch
				);
		return -1;
	}

	return i0 + (i1<<1);
}

// This is a self clocked RS232 bit stream generator, similar to manchester
// To send a 1 bit, it is low for 1 time period before going high.
// To send a 0 bit, it is low for 2 time periods before going high.
// This type of encoding is piggybacked on the RS232 protocol
// Thankfully RS232 Stop bits are always sent high, so they may be ignored
// Parity could be used to send one more bit occasionally, but
// changing that bit in a controlled way requires a system call and extra overhead.
// Therefore, for program speed/simplicity -- parity is none (high) and minimum
// stop bits is acceptable.

void Send32Bits( int port, u32 data ) {
	u8 txByte;
	u8 punch;
	int bits=32;
	
	if (!bits) return;
	while (bits>0) {
		punch=1; txByte=0xFF;
		while((1)) {
			if (!(data & 1)) {
				txByte ^= punch;
				punch<<=1;
			}
			data >>= 1; --bits;
			if(!bits || (punch==0x40 && !(data&1)) || !(punch&0x7F))
				break;  // Not enough room or data => break
			punch <<= 1; txByte ^= punch; punch <<=1;
		}
		ESendByte(port,txByte);
	}
}

// Alternate transmission algorithm to prove where the error lies
#if 0
static void Send32Bits(int port, u32 value) {
	int i;
	for(i=0; i<16; ++i) {
		Send2Bits( port, value & 0x3 );
		value >>= 2;
	}
}
#endif

void mdelay(u32 ms) {

#ifdef __linux__   /* Linux */

    usleep(ms*1000);

#else              /* windows */

    Sleep(ms);

#endif

}

int GetResponse( int port, int ms ) {
	int response;
	int i;
	int wordsInTime = (baudrate*ms)/11000;  //11 bits/word roughly,1000ms/s
	int pollRate    = (baudrate*20)/11000;
	
	if (diagnose){
		fprintf(stderr,"Words before timeout (%d ms)=%d\n", ms, wordsInTime );
		fprintf(stderr,"Poll rate =%d words/sample\n"  , pollRate    );
	}
	errno=0;
	do {
		ESendByte(port,0xB6);	// 10110110:_ -> 11111??:_
		for(i=0; i<pollRate; ++i) {
			ESendByte( port,0xFF );
			if (ByteReady(port)) break; 
		}
		wordsInTime -= i;
	} while( !ByteReady(port) && wordsInTime > 0 );
	
	response = ReceiveByte(port);
	if (verbose && errno) {
		perror("GetResponse()");
		return -1;
	}
	if ( ((response & 0xFC) != 0xFC) || (!(response&1) != !(response&2)) ){
		if (diagnose) 	fprintf(stderr,"Invalid response: %02X\n",
					(unsigned)response
				);
		return -1;
	}

	if (response<0) return response;
	return !!(response & 2);
}

int prop_connect(int port) {
  unsigned char lfsr[125];
  int count;
  int i;
  int temp;

  int version;
  
  if (reset_delay > 0) {
     if (verbose) {
        fprintf(stderr, "delaying before reset\n");
     }
     mdelay(reset_delay);
     if (verbose) {
        fprintf(stderr, "reset\n");
     }
  }

#ifdef DEBUG_RESET
  fprintf(stderr, "ready to %sreset - press enter to continue\n", (double_reset ? "double " : ""));
  while ((i = getchar()) != '\n') {
     ;
  }
#endif


  // BradC recommended 'double-reset' routine:
  if (double_reset) {
     SetDTR(port);
     mdelay(RESET_TIME);
     ClrDTR(port);
     mdelay(RESET_TIME);
  }
  SetDTR(port);
  mdelay(RESET_TIME);
  ClrDTR(port);

  ClearInput(port);
  mdelay(95);
 
  count=0;
  while(count!=251) {
		errno=0;
		count+=SendBuf( port, lfsrOut+count, 251-count );
		if (errno && errno != EAGAIN) {
			perror("prop_connect(), probe for propeller signature");
			return -1;
		}
  }
  count =0;
  while( count!=125 ) {
	errno=0;
  	count+=PollComport( port, lfsr+count, 125-count ); // Get the data
	if (errno && errno != EAGAIN) {
			perror("prop_connect(), read propeller signature");
			return -1;	
	}
  }

  if (strncmp( lfsr, lfsrIn, 125 )) {
	if (verbose) fprintf(stderr,"Invalid signature, Propeller not found\n");
	return -1;
  } 
   
  // Extract a version number.
  version = 0;
  for (i=0; i<8; i+=2 ) {
	version >>= 2;
	version += (Get2Unaligned(port)  << 6);
  }
  return version;
}

// The propeller requires 32bit words during programming cycle, so
// only whole numbers of 32bit words are accepted, and len is that number

void prop_upload(int port, u8* buffer, size_t len) { 
  int response;
  int count=len/4; // How many longs are there 
  int byteCount;   // what size does the binary believe it is?

  u32 bigEnd = ~(((u32)-1) >> 1);
  if (*((char*)((void*)&bigEnd)))
	fprintf(stderr,"Warning from payload.c:prop_upload():\n"
		"This program is untested with big endian processors\n"
		"Byte order loading from binary is assumed little endian\n" 
	        "Endianness may change by modifying #define BINARY_BIGEND\n"
	);

  if ( (len<32) || (byteCount = buffer[8] | (buffer[9] << 8))  !=  4*count) {
	fprintf(stderr,"Corrupt binary detected, Aborting load.\n");
 	return; 
  }

  // Begin actual load sequence, Tell propeller what is coming
  if (eeprom)	Send32Bits( port, CMD_PROGRUN );
  else		Send32Bits( port, CMD_LOADRUN );

  Send32Bits( port, count ); // Number of longs to load

  --buffer; // prepare to parse buffer a little faster
#ifndef BINARY_BIGEND
     do { // load little endian style 32 bit numbers
 	  u32 word32=	+(((u32)(*++buffer)))
			+(((u32)(*++buffer))<<8)
			+(((u32)(*++buffer))<<16)
			+(((u32)(*++buffer))<<24);
	  Send32Bits( port, word32 );
	mdelay(1);
     } while(--count);
#else
     do { // Load big endian style 32 bit numbers
	  u32 word32=	+((u32)(++*buffer)<<24)
			+((u32)(++*buffer)<<16)
			+((u32)(++*buffer)<<8)
			+((u32)(++*buffer));
	  Send32Bits( port, word32 );
     } while(--count);
#endif

  // Zeroing and Checksum can take up to 195ms to complete 
  // Timeouts are from [75 to 175]ms (cold to hot chip)
  // 275ms would be a fairly conservative timeout allow 2x reads before timeout
  // given a 30ms polling interval for the zeroing cycle

  if ( GetResponse( port, 300 ) ){ // Allow generous time for checksum + timeout
	fprintf(stderr,"Checksum error while uploading %s program\n",
		(eeprom)?"eeprom":"ram memory"
	);
  }
 
  if (!eeprom) return;
  if ( GetResponse( port, 5000 ) ) { // Up to 5 seconds time to program
	fprintf(stderr,"Propeller did not succeed in programming the eprom\n");
 	return; 
  }

  if ( GetResponse( port , 2000 ) ) { // Up to two seconds to read/verify data
	fprintf(stderr,"EEprom programming did not verify correctly");
  }
  return;
}

void cat_WriteByte(int port, u8 b) {
   if (diagnose > 1) {
      fprintf(stderr, "Sending byte 0x%02x\n", b);
   }
   if (b == 0xff) {
      SendByte(port, b);
      SendByte(port, 0);
   }
   else {
      SendByte(port, b);
   }
}

void cat_WriteLong(int port, unsigned long l) {
   int i;

   for (i = 0; i < 4; i++) {
      cat_WriteByte(port, l & 0xff);
      l >>= 8;
   }
}

void cat_WriteSync(int port, u8 cpu) {
   SendByte(port, 0xff);
   SendByte(port, cpu);
}

void cat_WritePage(int port, u8 *page, unsigned long addr, u8 cpu) {
   int i;

   if (diagnose) {
      fprintf(stderr, "Sending page %08lX\n", addr);
   }
   cat_WriteLong(port, (cpu << 24) | addr);
   cat_WriteLong(port, PAGE_SIZE);
   for (i = 0; i < PAGE_SIZE; i++) {
      cat_WriteByte(port, page[i]);
      if (INTERCHAR_DELAY > 0) {
         mdelay(INTERCHAR_DELAY);
      }
   }

}

int cat_ReadByte(int port, int cpu) {
   int result;
   int i;

   static int save_data = 0;

   if (save_data & 0x100) {
      if (save_data != 0x1ff) {
        result = save_data & 0xff;
        save_data = 0;
        if (diagnose > 1) {
           fprintf(stderr, "Read byte 0x%02x\n", result);
        }
        return result;
      }
   }
   
   while (1) {
      for (i = 0; i < 5; i++) {
         result = ReceiveByte(port);
         if (diagnose > 1) {
            fprintf(stderr, "Read result %d\n", result);
         }
         if (result != -1) {
            break;
         }
      }
      if (result < 0) {
         if (diagnose > 1) {
            fprintf(stderr, "Read result %d\n", result);
         }
         return result;
      }
      if (save_data == 0x1ff) {
        if (result == cpu) {
           if (diagnose > 1) {
              fprintf(stderr, "Read sync for cpu %d\n", cpu);
           }
           return -2;
        }
        if (result == 0) {
           result = 0xff;
           save_data = 0;
        }
        else {
           save_data = result | 0x100;
           result = 0xff;
        }
        if (diagnose > 1) {
           fprintf(stderr, "Read byte 0x%02x\n", result);
        }
        return result;
      }
      else {
         if (result == 0xff) {
            save_data = 0x1ff;
         }
         else {
            save_data = 0;
            if (diagnose > 1) {
               fprintf(stderr, "Read byte 0x%02x\n", result);
            }
            return result;
         }
      }
   }
}

int cat_ReadSync(int port, int cpu) {
   int result;
   int retries;
   retries = 0;
   while (1) {
      result = ReceiveByte(port);
      //if (result < 0) {
         //return result;
      //}
      if (result == 0xff) {
         result = ReceiveByte(port);
         if (result == cpu) {
            return 1;
         }
      }
      retries++;
      if (retries > MAX_ATTEMPTS) {
         return 0;
      }
      if (synctime > 0) {
         mdelay (synctime);
      }
   }
}

u8 cat_LRC(u8 *buffer, int size) {
   int i;
   u8 result = 0;

   for (i = 0; i < size; i++) {
      result ^= buffer[i];
   }
   return result;
}

void cat_upload(int port, char *fname, int cpu) {
  FILE *f;
  char full_name[MAX_LINELEN + 1];
  unsigned char buffer[PAGE_SIZE];
  unsigned int addr;
  u8 lrc;
  int result;
  int res;
  int i;
  int retries;

  safecpy(full_name, fname, MAX_LINELEN);
  if ((strstr(full_name, ".binary") == NULL)
  &&  (strstr(full_name, ".bin") == NULL)
  &&  (strstr(full_name, ".eeprom") == NULL)) {
     safecat(full_name, ".binary", MAX_LINELEN);
  }
  if (verbose) {
     fprintf(stderr, "Loading Propeller binary %s\n", full_name);
  }

  f = fopen(full_name, "rb");
  if (f == NULL) {
    fprintf(stderr, "Error: Unable to open file %s\n", full_name);
    return;
  }

  ClearInput(port);
  cat_WriteSync(port, cpu);
  addr = 0;
  while (1) {
     for (i = 0; i < PAGE_SIZE; i++) {
        buffer[i] = 0;
     }
     res = fread(buffer, sizeof(char), PAGE_SIZE, f);
     if (res > 0) {
        lrc = cat_LRC (buffer, PAGE_SIZE);
        retries = 0;
        while (retries < MAX_ATTEMPTS) {
           if ((retries > 0) && verbose) {           
              fprintf(stderr, "Retry %d\n", retries);
           }
           cat_WritePage(port, buffer, addr, cpu);
           mdelay (INTERPAGE_DELAY);
           if (verbose) {
              fprintf(stderr, ".");
           }
           if (cat_ReadSync(port, cpu)) {
              if ((result = cat_ReadByte(port, cpu)) >= 0) {
                 if (lrc == result) {
                    addr += PAGE_SIZE;
                    break;
                 }
                 else {
                    fprintf(stderr, "Warning: incorrect LRC for addr %08x - expected 0x%02x, received 0x%02x\n", addr, lrc, result);
                 }
              }
              else {
                 fprintf(stderr, "Warning: no LRC received for addr %08x\n", addr);
              }
           }
           else {
              fprintf(stderr, "Warning: no sync received for addr %08x\n", addr);
           }
           retries++;
        }
        if (retries == MAX_ATTEMPTS) {
           fprintf(stderr, "Error: too many retries for addr %08x\n", addr);
           return;
        }
     }
     else {
        break;
     }
  }
  cat_WriteLong(port, SIO_EOP | (cpu << 24));
  cat_WriteLong(port, 0);
  if (verbose) {
     fprintf(stderr, "\n");
  }
  
}


int prop_open(int port, int baudrate, int timeout, int verbose) {
  return OpenComport(port, baudrate, timeout, verbose);
  // flush input
  //while (ByteReady(port)) {
     //ReceiveByte(port);
  //}
}


void prop_close(int port) {
  CloseComport(port);
  // flush input
  //while (ByteReady(port)) {
     //ReceiveByte(port);
  //}
}



// Up to 32k in size (plus 1 so we detect files larger than 32k).
unsigned char buffer[MAX_PROGRAM_SIZE + 1];

int LoadPropellerFile(const char* fname) {
  char full_name[MAX_LINELEN + 1];
  FILE *f;
  int res;

  safecpy(full_name, fname, MAX_LINELEN);
  if ((strstr(full_name, ".binary") == NULL)
  &&  (strstr(full_name, ".bin") == NULL)
  &&  (strstr(full_name, ".eeprom") == NULL)) {
     safecat(full_name, ".binary", MAX_LINELEN);
  }
  if (verbose) {
     fprintf(stderr, "Loading Propeller binary %s\n", full_name);
  }

  f = fopen(full_name, "rb");
  if (f == NULL) {
    fprintf(stderr, "Error: Unable to open file %s\n", full_name);
    return 0;
  }
  res = fread(buffer, sizeof(char), sizeof(buffer)/sizeof(char), f);
  if (res == sizeof(buffer)) {
    fprintf(stderr, "Error: %s is larger than %d bytes - this is too large to load as the first file\n", fname, MAX_PROGRAM_SIZE);
    fclose(f);
    return 0;
  }
  if (res % 4) {
    fprintf(stderr, "Error: %s length (%d) is not a multiple of 4, - check it is a a Propeller binary\n", fname, res);
    fclose(f);
    return 0;
  }
  fclose(f);
  return res;
}

void help(char *my_name) {
   int i;

   fprintf(stderr, "usage: %s [options] propeller_file [catalina_file]\n\n", my_name);
   fprintf(stderr, "options:  -? or -h  print this helpful message and exit (-v for more help)\n");
   fprintf(stderr, "          -a port   find propeller port automatically, starting at port\n");
   fprintf(stderr, "          -b baud   use specified baudrate (default is %d)\n", DEFAULT_BAUDRATE);
   fprintf(stderr, "          -c cpu    cpu destination for catalina upload (default is %d)\n", DEFAULT_CPU);
   fprintf(stderr, "          -d        diagnostic mode (-d again for more diagnostics)\n");
   fprintf(stderr, "          -e        program the EEPROM with the program loaded\n");
   fprintf(stderr, "          -f msec   set interfile delay in milliseconds (default is %d)\n", DEFAULT_SYNCTIME);
   fprintf(stderr, "          -n msec   set sync timeout in milliseconds (default is %d)\n", DEFAULT_SYNCTIME);
   fprintf(stderr, "          -p port   use port for uploads (just first upload if -s used)\n");
   fprintf(stderr, "          -r msec   set reset delay in milliseconds (default is %d)\n", RESET_DELAY);
   fprintf(stderr, "          -s port   switch to port for second and subsequent uploads\n");
   fprintf(stderr, "          -t msec   set read timeout in milliseconds (default is %d)\n", DEFAULT_TIMEOUT);
   fprintf(stderr, "          -v        verbose mode (or include port numbers in help message)\n");
   fprintf(stderr, "          -w        wait for a keypress between each load\n");
   fprintf(stderr, "          -x        do catalina upload only (boot loader already loaded)\n");
   fprintf(stderr, "          -z        double reset\n");
   if (verbose) {
      fprintf(stderr, "\nport can be:\n");
      for (i = 0; i < ComportCount(); i++) {
         fprintf(stderr, "          %s%d = %s\n", (i < 9 ? " " : ""), i + 1, ShortportName(i));
      }
   }
}

/*
 * decode arguments, building file list - return -1 if
 * there is no further processing to do
 */
int decode_arguments (int argc, char *argv[]) {
   int  i = 0;
   int  code = 0;
   char filename[MAX_LINELEN + 3 + 1];

   while ((code >= 0) && (argc--)) {
      if (diagnose) {
         fprintf(stderr, "arg: %s\n", argv[i]);
      }
      if (i > 0) {
         if (argv[i][0] == '-') {
            if (diagnose) {
               fprintf(stderr, "switch: %s\n", argv[i]);
            }
            // it's a command line switch
            switch (argv[i][1]) {
               case 'h':
               case '?':
                  if (strlen(argv[0]) == 0) {
                     // in case my name was not passed in ...
                     help("payload");
                  }
                  else {
                     help(argv[0]);
                  }
                  code = -1;
                  break;
               case 'a':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &auto_port);
                     }
                     else {
                        fprintf(stderr, "Option -a requires a parameter\n");
                        code = -1;
                        break;
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &auto_port);
                  }
                  if (port != 0) {
                     fprintf(stderr, "Option -p overrides -a (-a ignored)\n");
                  }
                  else {
                     if ((auto_port < 1) || (auto_port > ComportCount())) {
                        fprintf(stderr, "Error: auto port must be between 1 and %d\n", ComportCount());
                        code = -1;
                        auto_port = 0;
                     }
                     else {
                        if (verbose) {
                           fprintf(stderr, "auto mode will start at port %s\n", ShortportName(auto_port-1));
                        }
                     }
                  }
                  break;
               case 'b':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &baudrate);
                     }
                     else {
                        fprintf(stderr, "Option -b requires a parameter\n");
                        code = -1;
                        break;
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &baudrate);
                  }
                  if ((baudrate < MIN_BAUDRATE) || (baudrate > MAX_BAUDRATE)) {
                     fprintf(stderr, "Error: baudrate must be in the range %d to %d\n", MIN_BAUDRATE, MAX_BAUDRATE);
                  }
                  else {
                     if (verbose) {
                        fprintf(stderr, "using baudrate %d\n", baudrate);
                     }
                  }
                  break;
               case 'c':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &cpu);
                     }
                     else {
                        fprintf(stderr, "Option -c requires a parameter\n");
                        code = -1;
                        break;
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &cpu);
                  }
                  if ((cpu < 1) || (cpu > MAX_CPU)) {
                     fprintf(stderr, "Error: cpu must be in the range 1 to %d\n", MAX_CPU);
                     code = -1;
                  }
                  else {
                     if (verbose) {
                        fprintf(stderr, "loading cpu %d\n", cpu);
                     }
                  }
                  break;
               case 'd':
                  diagnose++;   /* increase diagnosis level */
                  verbose = 1;   /* diagnose implies verbose */
                  if (diagnose == 1) {
                     fprintf(stderr, "Catalina Payload %s\n", VERSION); 
                  }
                  fprintf(stderr, "diagnostic level %d\n", diagnose);
                  break;
               case 'e':
                  eeprom = 1;
                  if (verbose) {
                     fprintf(stderr, "eeprom mode\n");
                  }
                  break;
               case 'f':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &interfile_delay);
                     }
                     else {
                        fprintf(stderr, "Option -f requires a parameter\n");
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &interfile_delay);
                  }
                  if ((interfile_delay < 0) || (interfile_delay > MAX_TIMEOUT)) {
                     fprintf(stderr, "Error: interfile delay must be in the range 0 to %d\n", MAX_TIMEOUT);
                     code = -1;
                  }
                  else {
                     if (verbose) {
                        fprintf(stderr, "using interfile delay %d\n", interfile_delay);
                     }
                  }
                  break;
                case 'n':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &synctime);
                     }
                     else {
                        fprintf(stderr, "Option -n requires a parameter\n");
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &synctime);
                  }
                  if ((synctime < 0) || (synctime > MAX_TIMEOUT)) {
                     fprintf(stderr, "Error: sync timeout must be in the range 0 to %d\n", MAX_TIMEOUT);
                     code = -1;
                  }
                  else {
                     if (verbose) {
                        fprintf(stderr, "using sync timeout %d\n", synctime);
                     }
                  }
                  break;
               case 'p':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &port);
                     }
                     else {
                        fprintf(stderr, "Option -p requires a parameter\n");
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &port);
                  }
                  if ((port < 1) || (port > ComportCount())) {
                     fprintf(stderr, "Error: port must be between 1 and %d\n", ComportCount());
                     code = -1;
                  }
                  else {
                     if ((auto_port > 1) && verbose) {
                        fprintf(stderr, "Option -p overrides -a\n");
                     }
                     auto_port = 0;
                     if (verbose) {
                        fprintf(stderr, "using port %s\n", ShortportName(port-1));
                     }
                  }
                  break;
               case 'r':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &reset_delay);
                     }
                     else {
                        fprintf(stderr, "Option -r requires a parameter\n");
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &reset_delay);
                  }
                  if ((reset_delay < 0) || (reset_delay > MAX_TIMEOUT)) {
                     fprintf(stderr, "Error: reset delay must be in the range 0 to %d\n", MAX_TIMEOUT);
                     code = -1;
                  }
                  else {
                     if (verbose) {
                        fprintf(stderr, "using reset_delay %d\n", reset_delay);
                     }
                  }
                  break;
               case 's':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &second_port);
                     }
                     else {
                        fprintf(stderr, "Option -s requires a parameter\n");
                        code = -1;
                        break;
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &second_port);
                  }
                  if ((second_port < 1) || (second_port > ComportCount())) {
                     fprintf(stderr, "Error: second port must be between 1 and %d\n", ComportCount());
                     code = -1;
                  }
                  else {
                     second_port--;
                     if (verbose) {
                        fprintf(stderr, "using %s as second port\n", ShortportName(second_port));
                     }
                  }
                  break;
               case 't':
                  if (strlen(argv[i]) == 2) {
                     // use next arg
                     if (argc > 0) {
                        sscanf(argv[++i], "%d", &timeout);
                     }
                     else {
                        fprintf(stderr, "Option -s requires a parameter\n");
                     }
                     argc--;
                  }
                  else {
                     // use remainder of this arg
                     sscanf(&argv[i][2], "%d", &timeout);
                  }
                  if ((timeout < 1) || (timeout > MAX_TIMEOUT)) {
                     fprintf(stderr, "Error: timeout must be in the range 1 to %d\n", MAX_TIMEOUT);
                     code = -1;
                  }
                  else {
                     if (verbose) {
                        fprintf(stderr, "using timeout %d\n", timeout);
                     }
                  }
                  break;
               case 'v':
                  verbose = 1;
                  if (diagnose == 0) {
                     fprintf(stderr, "Catalina Payload %s\n", VERSION); 
                  }
                  fprintf(stderr, "verbose mode\n");
                  break;
               case 'w':
                  wait = 1;
                  if (verbose) {
                     fprintf(stderr, "wait mode enabled\n");
                  }
                  break;
               case 'x':
                  xmm_only = 1;
                  fprintf(stderr, "xmm only mode (boot loader already loaded)\n");
                  break;
               case 'z':
                  double_reset = 1;
                  if (verbose) {
                     fprintf(stderr, "double reset enabled\n");
                  }
                  break;
               default:
                  fprintf(stderr, "\nError: unrecognized switch: %s\n", argv[i]);
                  code = -1; // force exit without further processing
                  break;
            }
         }
         else {
            // assume its a filename
            if (file_count < MAX_FILES) {
               file_name[file_count++] = strdup(argv[i]);
               code = 1; // work to do
            }
            else {
               fprintf(stderr, "\ntoo many files specified - file %s will be ignored\n", argv[i]);
            }
         }
      }
      i++; // next argument
   }
   if (code == -1) {
      return code;
   }
   if (diagnose) {
      fprintf(stderr, "executable name = %s\n", argv[0]);
   }
   if ((argc == 1) || (file_count == 0)) {
      if (!verbose) {
         fprintf(stderr, "Catalina Payload %s\n", VERSION); 
      }
      if (strlen(argv[0]) == 0) {
         // in case my name was not passed in ...
         help("payload");
      }
      else {
         help(argv[0]);
      }
      code = -1;
   }

   return code;

}



int main(int argc, char* argv[]) {
  int i;
  char c;
  int res = 0;
  int length = 0;
  int version = 0;

   if (decode_arguments(argc, argv) <= 0) {
      if (diagnose) {
         fprintf(stderr, "%s exiting\n", argv[0]);
      }
      exit(0);
   }

   if (file_count == 0) {
      fprintf(stderr, "\nNo input files specified - exiting\n");
      exit(1);
   }

   if (!xmm_only) {
      length = LoadPropellerFile(file_name[0]);
      if (!length) {
         exit(1);
      }
      if (diagnose) {
         fprintf(stderr, "File size = %d\n", length);
      }
   }

   if (diagnose) {
      fprintf(stderr, "baudrate = %d\n", baudrate);
      fprintf(stderr, "timeout = %d\n", timeout);
      fprintf(stderr, "sync timeout = %d\n", synctime);
      fprintf(stderr, "reset delay = %d\n", reset_delay);
      fprintf(stderr, "interfile delay = %d\n", interfile_delay);
   }
   if (auto_port) {
      port = auto_port - 1;
      while (port < ComportCount()) {
         if (diagnose) {
            fprintf(stderr, "Trying %s ... ", ShortportName(port));
         }
         res = prop_open(port, baudrate, timeout, diagnose);
         if (res == 0) {
            if (diagnose) {
               fprintf(stderr, "opened port\n");
            }
            if (xmm_only) {
               fprintf(stderr, "Using port %s for %supload\n", ShortportName(port), (file_count == 1 ? "" : "first "));
               break;
            }
            else {
               version = prop_connect(port);
               if (version > 0) {
                  fprintf(stderr, "Found Propeller (version %d) on port %s for %supload\n", version, ShortportName(port), (file_count == 1 ? "" : "first "));
                  break;
               }
               else {
                  prop_close(port);
                  res = -1;
               }
            }
         }
         else {
            prop_close(port);
            res = -1;
         }
         port++;
      }
      if (res != 0) {
         fprintf(stderr, "No Propeller found on any port\n");
         exit(1);
      }
   }
   else {
      if (port) {
         port--;
         res = prop_open(port, baudrate, timeout, 1);
         if (res) {
           fprintf(stderr, "Error: Unable to initialize Propeller - check connections\n");
           exit(1);
         }
         if (xmm_only) {
            fprintf(stderr, "Using port %s for %supload\n", ShortportName(port), (file_count == 1 ? "" : "first "));
         }
         else {
            version = prop_connect(port);
            if (version > 0) {
               fprintf(stderr, "Using Propeller (version %d) on port %s for %supload\n", version, ShortportName(port), (file_count == 1 ? "" : "first "));
            }
            else {
               fprintf(stderr, "No Propeller found on port %s\n", ShortportName(port));
               prop_close(port);
               exit(1);
            }
         }
      }
      else {
         fprintf(stderr, "No port specified, and auto port mode not enabled\n");
      }
 
   }
 
   if (xmm_only || (version == 1)) {

      if (!xmm_only) {
         // Upload a normal (<=32kb) binary file to the Propeller using the Parallax protocol
         if (verbose) {
           fprintf(stderr, "Uploading normal binary file %s\n", file_name[0]);
         }
         prop_upload(port, buffer, length);
	if (verbose) fprintf(stderr,"Completed upload\n");
      }
      // if more files are specifed, upload subsequent files using the Catalina protocol
      for (i = (xmm_only ? 0 : 1); i < file_count; i++) {
         if (wait) {
            fprintf(stderr, "\nWaiting before loading next file - press ENTER to proceed ...\n");
            while ((c = getchar()) != '\n') {
               ;
            }
         }
         else {
            if (interfile_delay > 0) {
               mdelay (interfile_delay);
            }
         }
         if ((second_port == -1) && (i > 0)) {
            fprintf(stderr, "Using port %s for subsequent uploads\n", ShortportName(port));
         }
         if ((port != second_port)  && (second_port != -1)) {
            // switch to second port for all subsequent files
            prop_close(port);
            port = second_port;
            res = prop_open(port, baudrate, timeout, diagnose);
            if (res == 0) {
               fprintf(stderr, "Using port %s for subsequent uploads\n", ShortportName(port));
            }
            else {
               fprintf(stderr, "Error: cannot open second port\n");
               exit(1);
            }
         }
         if (verbose) {
            fprintf(stderr, "Uploading Catalina binary file %s\n", file_name[i]);
         }
         cat_upload(port, file_name[i], cpu);
         prop_close(port);
      }
   }
   else {
     fprintf(stderr, "Error: Unknown Propeller version\n");
   }
   
   return 0;
}


